from pydantic import BaseModel

from . import exceptions
from .classy_config import ClassyConfig
from .config_param import ConfigParam

__all__ = [
    "exceptions",
    "ClassyConfig",
    "ConfigParam"
]



import json
from os import PathLike
from typing import Optional, Union

from .exceptions import DoubleCreation


class ClassyConfig:
    """Define your config filepath."""

    instance: Optional["ClassyConfig"] = None
    _raw_config: Optional[dict] = None

    def __new__(cls, *args, **kwargs) -> "ClassyConfig":
        """
        Singleton implementation.

        if the object has already been created a DoubleCreation error is raised.
        """
        if cls.instance is not None:
            raise DoubleCreation("Can only create one ClassyConfig")

        cls.instance = super(ClassyConfig, cls).__new__(cls)
        return cls.instance

    def __init__(self, config_file: Union[str, PathLike]) -> None:
        self.config_file = config_file

    @property
    def raw_config(self) -> dict:
        """
        Gathers the raw config from the file.

        This config file is read once & saved to an instance attribute,
            so future access will not require another file read.
        """
        if self._raw_config is None:
            with open(self.config_file, "r") as f:
                self._raw_config = json.load(f)

        return self._raw_config



class ClassyConfigException(BaseException):
    """
    All ClassyConfig exceptions inherit from this exception.

    This is to allow a generic exception handler for
    any exception that could be raised from this module.
    """


class DoubleCreation(ClassyConfigException):
    """Raised when attempting to create the global context for the second time."""


class InstanceNotCreated(ClassyConfigException):
    """Raised when attempting to get a ConfigParam before creating the global context."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['classy_config']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0', 'typing-inspect>=0.7.1,<0.8.0']

setup_kwargs = {
    'name': 'classy-config',
    'version': '0.1.8',
    'description': 'Inject your config variables into methods, so they are as close to usage as possible.',
    'long_description': '# ClassyConfig\n\n`ClassyConfig` is a Python3 package aiming to remove the need for a `config.py` or `settings.py` file.\n\n```py\n\nfrom classy_config import BaseModel, ClassyConfig, ConfigParam\n\n# Create your global config manager (example test-config.json below)\nconfig = ClassyConfig(config_file="test-config.json")\n\n# Resolve default values based on your config\ndef print_current_version(version: str = ConfigParam("version", str)) -> None:\n    print(version)\n\n# Use Pydantic Models for your config\nclass Author(BaseModel):\n    username: str\n    email: str\n    lucky_number: int\n\n# Resolve default values based on your config\ndef print_author(author: Author = ConfigParam("author", Author)) -> None:\n    print(author)\n    \n# Allows for nested values\ndef print_value(value: int = ConfigParam("nested.value", int)) -> None:\n    print(value)\n```\n```json\n{\n  "version": "0.0.1",\n  \n  "author": {\n    "username": "GDWR",\n    "email": "gregory.dwr@gmail.com",\n    "lucky_number": 17\n  },\n\n  "nested": {\n    "value": 10\n  }\n}\n```\n',
    'author': 'GDWR',
    'author_email': 'gregory.dwr@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/GDWR/classy-config',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from typing import Dict, List, Type, TypeVar

from pydantic import BaseModel
from typing_inspect import get_origin

from .classy_config import ClassyConfig
from .exceptions import InstanceNotCreated

T = TypeVar("T")


class _ResolveFromConfig(type):
    def __call__(cls, variable_path: str, _type: Type[T], deliminator: str = ".") -> T:

        config = ClassyConfig.instance
        if config is None:
            raise InstanceNotCreated("Please create an instance of ClassyConfig before using ConfigParam")

        data = config.raw_config
        for step in variable_path.split(deliminator):
            try:
                data = data[step]
            except KeyError:
                raise KeyError(f"Config: {variable_path} does not exist")

        if get_origin(_type) in (List, Dict):
            return data
        if issubclass(_type, BaseModel):
            return _type(**data)
        else:
            return _type(data)


class ConfigParam(metaclass=_ResolveFromConfig):
    """Resolve to a value defined within the config."""
